/* protocols.c
	Suuport functions for sendip.c
*/

#define __USE_BSD     /* GLIBC */
#define _BSD_SOURCE   /* LIBC5 */
#define __FAVOR_BSD   /* Some of GLIBC */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <stdlib.h>
#include "csum.h"

#ifndef __GNU_LIBRARY__
#define __GNU_LIBRARY__ 42
#endif
#if __GNU_LIBRARY__ < 6
#define icmp     icmphdr
#define icmp_type type
#define icmp_code code
#define icmp_cksum checksum
#define uh_sport source
#define uh_dport dest
#define uh_ulen  len
#define uh_sum   check
#endif

/* Handy macros to set the TCP flags */
#define SETRES(tcp,i) tcp->th_flags &= (i)<<6;
#define SETURG(tcp,i) (i)?(tcp->th_flags|=TH_URG):(tcp->th_flags&=~TH_URG);
#define SETACK(tcp,i) (i)?(tcp->th_flags|=TH_ACK):(tcp->th_flags&=~TH_ACK);
#define SETPSH(tcp,i) (i)?(tcp->th_flags|=TH_PUSH):(tcp->th_flags&=~TH_PUSH);
#define SETRST(tcp,i) (i)?(tcp->th_flags|=TH_RST):(tcp->th_flags&=~TH_RST);
#define SETFIN(tcp,i) (i)?(tcp->th_flags|=TH_FIN):(tcp->th_flags&=~TH_FIN);
#define SETSYN(tcp,i) (i)?(tcp->th_flags|=TH_SYN):(tcp->th_flags&=~TH_SYN);

extern void usage_exit(const char *err);

void ip_initialize(struct ip *iph);
void icmp_initialize(struct icmp *icp);
void tcp_initialize(struct tcphdr *tcp);
void udp_initialize(struct udphdr *udp);

void do_ipopt(struct ip *iph, char *opt, char *arg);
void do_icmpopt(struct icmp *icp, char *opt, char *arg);
void do_tcpopt(struct tcphdr *tcp, char *opt, char *arg);
void do_udpopt(struct udphdr *udp, char *opt, char *arg);

void ip_finalize(struct ip *iph, int length, struct sockaddr_in *to);
void icmp_finalize(struct ip *iph, struct icmp *icp, int length);
void tcp_finalize(struct ip *iph, struct tcphdr *tcp, int length);
void udp_finalize(struct ip *iph, struct udphdr *udp, int length);

void ip_initialize(struct ip *iph) {
	iph->ip_v=     (unsigned int)0x4;
	iph->ip_hl=    (unsigned int)0x5;
	iph->ip_tos=       (u_int8_t)0x00;
	iph->ip_len=      (u_int16_t)0xFFFF;        /* automatic */
	iph->ip_id=       (u_int16_t)rand();
	iph->ip_off=      (u_int16_t)0x0000;
	iph->ip_ttl=       (u_int8_t)0xFF;
	iph->ip_p=         (u_int8_t)0xFF;          /* automatic */
	iph->ip_sum=      (u_int16_t)0xFFFF;        /* automatic */
	iph->ip_src.s_addr=          INADDR_NONE;   /* automatic */
	iph->ip_dst.s_addr=          INADDR_NONE;   /* automatic */
}

void icmp_initialize(struct icmp *icp) {
	icp->icmp_type=              ICMP_ECHO;
	icp->icmp_code=    (u_int8_t)0x00;
	icp->icmp_cksum=  (u_int16_t)0xFFFF;        /* automatic */
}

void tcp_initialize(struct tcphdr *tcp) {
	tcp->th_sport =   (u_int16_t)0x0000;
	tcp->th_dport =   (u_int16_t)0x0000;
	tcp->th_seq =     (u_int32_t)rand();
	tcp->th_ack =     (u_int32_t)0x00000000;
	tcp->th_off =     (u_int16_t)0x5;
	tcp->th_x2 =      (u_int16_t)0x0;
	tcp->th_flags =   (u_int16_t)TH_SYN;
	tcp->th_win =     (u_int16_t)65535;
	tcp->th_sum =     (u_int16_t)0xFFFF;  /* automatic */
	tcp->th_urp =     (u_int16_t)0x0000;
}

void udp_initialize(struct udphdr *udp) {
	udp->uh_sport =     (u_int16_t)0x0000;
	udp->uh_dport =       (u_int16_t)0x0000;
	udp->uh_ulen =        (u_int16_t)0xFFFF;        /* automatic */
	udp->uh_sum =      (u_int16_t)0xFFFF;        /* automatic */
}

void do_ipopt(struct ip *iph, char *opt, char *arg) {
	char rb[32];
	if(*arg == 'r') {
		sprintf(rb,"%d",rand());
		arg = rb;
	}
	switch(opt[2]) {
	case 'h':
		iph->ip_hl = (unsigned int)strtoul(arg, (char **)NULL, 0) & 0xF;
		break;
	case 'y':
		iph->ip_tos = (u_int8_t)strtoul(arg, (char **)NULL, 0);
		break;
	case 'l':
		iph->ip_len = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 'i':
		iph->ip_id = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 't':
		iph->ip_ttl = (u_int8_t)strtoul(arg, (char **)NULL, 0);
		break;
	case 'p':
	   iph->ip_p = (u_int8_t)strtoul(arg, (char **)NULL, 0);
		break;
	case 'c':
		iph->ip_sum = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 's':
		iph->ip_src.s_addr = inet_addr(arg);
		break;
	case 'd':
		iph->ip_dst.s_addr = inet_addr(arg);
		break;
	case 'f':
		switch(opt[3]) {
		case 'r':
			if(arg != rb && *arg != '0' && *arg != '1') {
				usage_exit("IP reserved flag must be 0 or 1\n");
			}
			*arg-='0';
			if(*arg) {
				iph->ip_off |= (u_int16_t)0x8000;
			} else {
				iph->ip_off &= (u_int16_t)0x7FFF;
			}
			break;
		case 'd':
			if(arg != rb && *arg != '0' && *arg != '1') {
				usage_exit("IP don't fragment flag must be 0 or 1\n");
			}
			*arg-='0';
			if(*arg) {
				iph->ip_off |= (u_int16_t)0x4000;
			} else {
				iph->ip_off &= (u_int16_t)0xBFFF;
			}
			break;
		case 'm':
			if(arg != rb && *arg != '0' && *arg != '1') {
				usage_exit("IP more fragments flag must be 0 or 1\n");
			}
			*arg-='0';
			if(*arg) {
				iph->ip_off |= (u_int16_t)0x2000;
			} else {
				iph->ip_off &= (u_int16_t)0xDFFF;
			}
			break;
		case '\0':
			iph->ip_off &= (u_int16_t)0xE000;
			iph->ip_off |= (u_int16_t)strtoul(arg, (char **)NULL, 0) & 
				(u_int16_t)0x1FFF;
			break;
		default:
			usage_exit("unknown IP flag\n");
			break;
		}
		break;
	default:
		usage_exit("unknown IP option\n");
		break;
	}
}

void do_icmpopt(struct icmp *icp, char *opt, char *arg) {
	char rb[32];
	if(*arg == 'r') {
		sprintf(rb,"%d",rand());
		arg = rb;
	}	switch(opt[2]) {
	case 't':
		icp->icmp_type = (u_int8_t)strtoul(arg, (char **)NULL, 0);
		break;
	case 'd':
		icp->icmp_code = (u_int8_t)strtoul(arg, (char **)NULL, 0);
		break;
	case 'c':
		icp->icmp_cksum = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	default:
		usage_exit("unknown ICMP option\n");
		break;
	}
}

void do_tcpopt(struct tcphdr *tcp, char *opt, char *arg) {
	char rb[32];
	if(*arg == 'r') {
		sprintf(rb,"%d",rand());
		arg = rb;
	}
	switch(opt[2]) {
	case 's':
		tcp->th_sport = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 'd':
		tcp->th_dport = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 'n':
		tcp->th_seq = htonl((u_int32_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 'a':
		tcp->th_ack = htonl((u_int32_t)strtoul(arg, (char **)NULL, 0));
		SETACK(tcp,1);
		break;
	case 't':
		tcp->th_off = (u_int16_t)strtoul(arg, (char **)NULL, 0) &0xF;
		break;
	case 'r':
		tcp->th_x2 = (u_int16_t)(strtoul(arg, (char **)NULL, 0) & 0x003C)>>2;
		SETRES(tcp,(u_int16_t)strtoul(arg, (char **)NULL, 0) & 0x0003);
		break;
	case 'f':
		if(arg != rb && *arg != '0' && *arg != '1') {
			usage_exit("TCP flag must be 0 or 1\n");
		}
		*arg-=0x30;
		switch(opt[3]) {
		case 'u':
			SETURG(tcp,(u_int16_t)*arg&1);
			break;
		case 'a':
			SETACK(tcp,(u_int16_t)*arg&1);
			break;
		case 'p':
			SETPSH(tcp,(u_int16_t)*arg&1);
			break;
		case 'r':
			SETRST(tcp,(u_int16_t)*arg&1);
			break;
		case 's':
			SETSYN(tcp,(u_int16_t)*arg&1);
			break;
		case 'f':
			SETFIN(tcp,(u_int16_t)*arg&1);
			break;
		default:
			usage_exit("TCP flage not known\n");
			break;
		}
		break;
	case 'w':
		tcp->th_win = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 'c':
		tcp->th_sum = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 'u':
		tcp->th_urp = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		SETURG(tcp,1);
		break;
	default:
		usage_exit("unknown TCP options\n");
		break;
	}
}

void do_udpopt(struct udphdr *udp, char *opt, char *arg) {
	char rb[32];
	if(*arg == 'r') {
		sprintf(rb,"%d",rand());
		arg = rb;
	}
	switch(opt[2]) {
	case 's':
		udp->uh_sport = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 'd':
		udp->uh_dport = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 'l':
		udp->uh_ulen = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	case 'c':
		udp->uh_sum = htons((u_int16_t)strtoul(arg, (char **)NULL, 0));
		break;
	default:
		usage_exit("unknown UDP options\n");
		break;
	}
}

void ip_finalize(struct ip *iph, int length, struct sockaddr_in *to) {
	if(iph->ip_len==(u_int16_t)0xFFFF) iph->ip_len = htons(length);
	if(iph->ip_p == (u_int8_t)0xFF) iph->ip_p=0;
	if(iph->ip_sum == (u_int16_t)0xFFFF) {
		iph->ip_sum = 0;
		iph->ip_sum = csum((unsigned short *)iph, sizeof(struct ip));
	}
	if(iph->ip_src.s_addr == INADDR_NONE)
		iph->ip_src.s_addr = inet_addr("127.0.0.1");
	if(iph->ip_dst.s_addr == INADDR_NONE)
		iph->ip_dst.s_addr = to->sin_addr.s_addr;
}

void icmp_finalize(struct ip *iph, struct icmp *icp, int length) {
	if(iph->ip_p == (u_int8_t)0xFF) iph->ip_p = IPPROTO_ICMP;
	if(icp->icmp_cksum == (u_int16_t)0xFFFF) {
		icp->icmp_cksum = 0;
		icp->icmp_cksum = csum((unsigned short *)icp,
									length-sizeof(struct ip));
	}
}

void tcp_finalize(struct ip *iph, struct tcphdr *tcp, int length) {
	if(iph->ip_p == (u_int8_t)0xFF) iph->ip_p = IPPROTO_TCP;
	if(tcp->th_sum == (u_int16_t)0xFFFF) {
		tcp->th_sum = 0;
		tcp->th_sum = tcpcsum((unsigned char *)iph, (unsigned short *)tcp,
								length-sizeof(struct ip));
	}
}

void udp_finalize(struct ip *iph, struct udphdr *udp, int length) {
	if(udp->uh_ulen==(u_int16_t)0xFFFF)
		udp->uh_ulen = htons(length-sizeof(struct ip));
	if(iph->ip_p == (u_int8_t)0xFF) iph->ip_p = IPPROTO_UDP;
	if(udp->uh_sum == (u_int16_t)0xFFFF) {
		udp->uh_sum = 0;
		udp->uh_sum = udpcsum((unsigned char *)iph, (unsigned short *)udp,
									length-sizeof(struct ip));
	}
}
